function [beta, bias, finalObj, changeLabel] = trcl_mkmtfl_md(trInd,C,pk)
%function [beta, bias, obj] = trcl_mkmtfl_md(trInd,C,pk)
%Implements a mirror descent algorithm for performing
%           multiple kernel, multi-task feature learning (classification)
%Input:
%GLOBAL VARIABLES:
%K - 3-d matrix such that K(:,:,j) is the gram-matrix of examples
%    from all the tasks with the jth kernel
%y - y is a column vector of labels of all examples across tasks
%ARGUMENTS:
%trInd - Tx1 cell such that trInd{t} is a column vector of training indices into
%          K and y
%C - vector of regularization parameters
%pk - vector of norms over kernels pk \in [1,2)
%OUTPUT:
%beta - 4-d cell of learnt coefficients. beta{k,c,t,j} is a column vector
%       of learnt coeff. 'k', 'c', 't', 'j' are indices into
%       pk,C,1:T,1:k vectors.
%bias - 3-d cell of bias values.
%finalObj - The optimal objective value
%changeLabel - as libsvm sometimes internally flips the labels, book keeping
%     is done in this variable in order to get correct output in predictions
fprintf('trcl_mkmtfl_md is being called\n');

global K
global y

if(any(pk<=1) || any(pk>=2))
    fprintf('Invalid pk. Working with truncated set\n');
    pk(find(pk<=1)) = 1.0001;
    pk(find(pk>=2)) = 1.9999;
end
myEps = 10e-4;
ITER_MAX = 200;
T = size(trInd,1);
totTrInd = [];
for t=1:T
    totTrInd = [totTrInd;trInd{t}];
end
m = length(totTrInd);
k = size(K,3);
M = cell(T,k);
delta = (1e-8)/k;
for t=1:T
    mt(t) = size(trInd{t},1);    %No. examples per task
    mpos = sum(y(trInd{t})==1);
    lib_para_base{t}=sprintf('-q -t 4 -w1 %f -w-1 %f', (mt(t)-mpos)/mt(t), mpos/mt(t));
end

%Kernelization tricks
for j=1:k
    [U{j},D] = eig((K(totTrInd,totTrInd,j)+K(totTrInd,totTrInd,j)')/2);
    rankQ(j,1) = rank(D);
    tmpnzpo = m-rankQ(j)+1;
    U{j} = D(tmpnzpo:m,tmpnzpo:m)^(-0.5)*U{j}(:,tmpnzpo:m)';
    for t=1:T
        M{t,j} = U{j}*K(totTrInd,trInd{t},j);
    end
end

%Loop for each parameter value
CLen = length(C); pkLen = length(pk);
for pkInd = 1:pkLen
    fprintf('pkInd number: %d\n',pk(pkInd));
    qbar = pk(pkInd)/(2-pk(pkInd));
    for CInd = 1:CLen
        fprintf('C = %d\n',C(CInd));
        for j=1:k
            L_Q{j} = eye(rankQ(j));
            lambda_Q{j} = 1/(k^(1/qbar)*rankQ(j))*ones(rankQ(j),1);
        end
        isConv = 0;
        for t=1:T
            lib_para{t} = sprintf('%s -c %f',lib_para_base{t},C(CInd));
        end
        %%SOLVER%%
        obj = realmax;
        %Start of Mirror Descent
        for iter = 1:ITER_MAX
            fprintf('.');
            obj_old = obj;
            %Compute gradient g_g{j} by solving T SVMs with current Q{j}
            obj=0;
            %computing Q{j} (aliased by g_g{j})
            for j=1:k
                g_g{j} = L_Q{j}*diag(lambda_Q{j})*L_Q{j}';
            end
            %For each task solve SVM
            for t=1:T
                %Compute temp. kernel
                tmpKern = zeros(mt(t));
                for j=1:k
                    tmpKern = tmpKern + M{t,j}'*g_g{j}*M{t,j};
                end
                %Compute alpha
                models{t} = svmtrain(y(trInd{t}),[(1:mt(t))' tmpKern],lib_para{t});
                %Compute Obj
                obj = obj + sum(abs(models{t}.sv_coef)) -0.5*models{t}.sv_coef'*tmpKern(models{t}.SVs,models{t}.SVs)*models{t}.sv_coef;
            end
            %Check for convergence of md
%                 fprintf('At iter=%d: obj=%f\n',iter,obj);
            if(obj_old < obj)
                    fprintf('Something wrong with convergence! obj_old=%f, obj=%f\n',obj_old,obj);
            elseif((obj_old-obj)/obj_old < myEps)
                finalObj{pkInd,CInd} = obj;
                isConv=1;
                break;
            end    
            if(iter == ITER_MAX)
                break;
            end
            %Prepare gradient g_g
            for j=1:k
                g_g{j} = zeros(rankQ(j));
                for t=1:T
                    tmpVec = M{t,j}(:,models{t}.SVs)*models{t}.sv_coef;
                    g_g{j} = g_g{j} + tmpVec*tmpVec';
                end
                g_g{j} = -0.5*g_g{j};
            end
            %End of computing gradient g_g{j} by solving T SVRs
            %Compute P{j} matrices (aliased by g_g{j} itself)
            %00-scatten norm for step-size computation
            eigsopts.disp = 0;
            for j=1:k
                g_ginf(j) = abs(eigs((g_g{j}+g_g{j}')/2,1,'lm',eigsopts));
            end
            s_g = sqrt(log(sum(rankQ))/iter)/max(g_ginf);                     %%%%%%%%%Can be avoided if expensive
            mu=[];
            for j=1:k
                g_g{j} = s_g*g_g{j}-(L_Q{j}*diag(log(lambda_Q{j}+delta)+1)*L_Q{j}');
                [L_Q{j},g_g{j}] = eig((g_g{j}+g_g{j}')/2);g_g{j} = diag(g_g{j});
                mu = [mu;g_g{j}];
            end 
            lambda = solveGrVarProb(mu,rankQ,qbar);
            for j=1:k
                lambda_Q{j} = lambda(sum(rankQ(1:j-1))+1:sum(rankQ(1:j-1))+rankQ(j));
            end
        end
        %End of Mirror Descent
        %%END OF SOLVER%%
        %Update beta
        for j=1:k
            g_g{j} = U{j}'*L_Q{j}*diag(lambda_Q{j})*L_Q{j}';
            for t=1:T
                beta{pkInd,CInd,t,j} = g_g{j}*M{t,j}(:,models{t}.SVs)*models{t}.sv_coef;
            end
        end
	for t=1:T
	    bias{pkInd,CInd,t} = models{t}.rho;
        changeLabel(t,1) = models{t}.Label(1);
	end
        if(isConv==1)
            fprintf('Successfull convergence\n');
        else
            fprintf('Convergence premature\n');
        end
    %End loop for C 
    end
%End loop for pk
end
% fprintf('iterations f=%d, g=%d\n',iter_f,iter_g);
